// Copyright (c) 2015 HigPen Works (Kenichiro Matsuura / Yuki Tsukasa). All rights reserved.
#pragma once
#include <cmath>
#include <GLKit/GLKit.h>
using namespace std;

static const float PI=(float)M_PI;

class CColor {
	GLKVector4& V() const { return *(GLKVector4*)this; }
public:
	float R, G, B, A;
	CColor(const GLKVector4& v) { V()=v; }
	operator GLKVector4&() { return V(); }
	CColor() : R(0), G(0), B(0), A(0) {}
    CColor(float r, float g, float b, float a) : R(r), G(g), B(b), A(a) {}
	bool operator==(const CColor& c) { return R==c.R && G==c.G && B==c.B && A==c.A; }
	CColor operator*=(const CColor& c) { return *this=GLKVector4Multiply(V(), c.V()); }
	friend CColor operator*(CColor a, const CColor& b) { return GLKVector4Multiply(a.V(), b.V()); }
};

class CVector {
	friend class CQuaternion;
	GLKVector3& V() const { return *(GLKVector3*)this; }
	float Length() const { return GLKVector3Length(V()); }
 	CVector Normalize() const { return GLKVector3Normalize(V()); }
	float Dot(const CVector& b) const { return GLKVector3DotProduct(V(), b.V()); }
	CVector Cross(const CVector& b) const { return GLKVector3CrossProduct(V(), b.V()); }
public:
	float X, Y, Z;
	CVector(const GLKVector3& v) { V()=v; }
	operator GLKVector3&() { return V(); }
	CVector() : X(0), Y(0), Z(0) {}
    CVector(float f) : X(f), Y(f), Z(f) {}
    CVector(float a, float b) : X(a), Y(b), Z(0) {}
    CVector(float a, float b, float c) : X(a), Y(b), Z(c) {}
	GLKMatrix4 MatrixPosition() const { return GLKMatrix4MakeTranslation(X, Y, Z); }
	GLKMatrix4 MatrixScale() const { return GLKMatrix4MakeScale(X, Y, Z); }
	bool operator==(const CVector& v) { return X==v.X && Y==v.Y && Z==v.Z; }
	CVector operator-() const { return CVector(-X, -Y, -Z); }
	CVector& operator+=(const CVector& v) { return V()=GLKVector3Add(V(), v.V()), *this; }
	CVector& operator-=(const CVector& v) { return V()=GLKVector3Subtract(V(), v.V()), *this; }
	CVector& operator*=(const CVector& v) { return V()=GLKVector3Multiply(V(), v.V()), *this; }
	CVector& operator/=(const CVector& v) { return V()=GLKVector3Divide(V(), v.V()), *this; }
	friend CVector operator+(const CVector& a, const CVector& b) { return GLKVector3Add(a.V(), b.V()); }
    friend CVector operator-(const CVector& a, const CVector& b) { return GLKVector3Subtract(a.V(), b.V()); }
    friend CVector operator*(const CVector& a, const CVector& b) { return GLKVector3Multiply(a.V(), b.V()); }
	friend CVector operator/(const CVector& a, const CVector& b) { return GLKVector3Divide(a.V(), b.V()); }
	friend float Length(const CVector& v) { return v.Length(); }
	friend CVector Normalize(const CVector& v) { return v.Normalize(); }
	friend float Dot(const CVector& a, const CVector& b) { return a.Dot(b); }
	friend CVector Cross(const CVector& a, const CVector& b) { return a.Cross(b); }
	CVector& Homing(const CVector& b, float f) {
		CVector v=b-*this;
		if (v.Length()<f) *this=b; else *this+=v.Normalize()*f;
        return *this;
	}
};

class CQuaternion {
	GLKQuaternion& Q() const { return *(GLKQuaternion*)this; }
	float X, Y, Z, W;
public:
	CQuaternion(const GLKQuaternion& q) { Q()=q; }
	operator GLKQuaternion&() { return Q(); }
    CQuaternion() : X(0), Y(0), Z(0), W(1) {}
	CQuaternion(const CVector& axis, float angle) { Q()=GLKQuaternionMakeWithAngleAndVector3Axis(angle*PI*2, Normalize(axis)); }
 	CQuaternion(const CVector& from, const CVector& to, const CVector& axis) {
		CVector fn=Normalize(from), tn=Normalize(to), a=Cross(fn, tn);
		float la=Length(a), d=Dot(fn, tn);
		if (la) Q()=GLKQuaternionMakeWithAngleAndVector3Axis(atan2f(la, d), Normalize(a));
		else Q()=GLKQuaternionMakeWithAngleAndVector3Axis(d>0?0:PI, Normalize(axis));
	}
	GLKMatrix4 MatrixRotation() const {
		CVector a=Axis();
		return Length(a)>0?GLKMatrix4MakeRotation(Angle()*PI*2, a.X, a.Y, a.Z):GLKMatrix4Identity;
	}
	CQuaternion& operator*=(const CQuaternion& q) { return Q()=GLKQuaternionMultiply(Q(), q.Q()), *this; }
	friend CQuaternion operator*(const CQuaternion& a, const CQuaternion& b) { return GLKQuaternionMultiply(a.Q(), b.Q()); }
	friend CVector operator*(const CVector& v, const CQuaternion& q) { return GLKQuaternionRotateVector3(q.Q(), v.V()); }
	CVector Axis() const { return GLKQuaternionAxis(Q()); }
	float Angle() const { return GLKQuaternionAngle(Q())/PI/2; }
	CQuaternion Inverse() const { return GLKQuaternionInvert(Q()); }
} __attribute__((aligned(16)));

static const CVector AXIS_X(1, 0, 0), AXIS_Y(0, 1, 0), AXIS_Z(0, 0, 1);
inline float Sign(float x) { return float(x>0?1:x<0?-1:0); }
inline CQuaternion RotationZ(float angle) { return CQuaternion(AXIS_Z, angle); }
inline float Sin(float x) { return sin(x*PI*2); }
inline float Cos(float x) { return cos(x*PI*2); }
inline float Tan(float x) { return tan(x*PI*2); }
