// Copyright (c) 2015 HigPen Works (Kenichiro Matsuura / Yuki Tsukasa). All rights reserved.
#pragma once
#include "Math.h"
#include "Texture.h"
#include <cassert>
using namespace std;

class CMover {
	friend class CMoverList;
	CMover *Prev, *Next;
public:
	CVector Position, Velocity, Scale;
	CQuaternion Rotation;
	CColor Color;
	CTexture* Texture;
	int State, Type, Time, Life;
	CMover* Owner;

	CMover() : Scale(1, 1, 1), Color(1, 1, 1, 1), Texture(0), State(0), Type(0), Time(0), Life(1), Owner(0) {}
	virtual ~CMover() {}

	virtual void Reset() {}
	virtual void Move() {}
	virtual void Draw();
};

class CMoverList {
	CMover *Active, *Free;
	int Size;
public:
	const int ActiveCount, FreeCount;
	template<class M> M* New() {
		assert(sizeof(M)<=Size);
		M *n=(M*)Free->Next, *m=(n==Free)?new(new char[Size]) M:(Free->Next=n->Next, new(n) M);
		m->Prev=Active->Prev, m->Next=Active, m->Prev->Next=m->Next->Prev=m;
		((int&)ActiveCount)++;
		return m;
	}
	void Delete(CMover* m) {
		m->~CMover();
		m->Prev->Next=m->Next, m->Next->Prev=m->Prev;
		m->Next=Free->Next, Free->Next=m;
		((int&)ActiveCount)--, ((int&)FreeCount)++;
	}
	CMoverList(int size) : Size(size), ActiveCount(0), FreeCount(0) {
		Active=new CMover(), Active->Prev=Active->Next=Active;
		Free=new CMover(), Free->Next=Free;
	}
	~CMoverList() {
		for (CMover *m=Active, *n; n=m->Next, delete m, m=n, m!=Active; ) ;
		for (CMover *m=Free, *n; n=m->Next, delete m, m=n, m!=Free; ) ;
	}
	void Move() {
		for (CMover *m=Active->Next; m!=Active; m=m->Next) if (m->Life>0) m->Move();
		for (CMover *m=Active->Next, *n; n=m->Next, m!=Active; m=n) if (m->Life<=0) Delete(m);
	}
	void Draw() {
		for (CMover* m=Active->Next; m!=Active; m=m->Next) if (m->Life>0) m->Draw();		
	}
	void Kill() {
		for (CMover* m=Active->Next; m!=Active; m=m->Next) m->Life=0;
	}
	template<class F> void Apply(F func) {
		for (CMover* m=Active->Next; m!=Active; m=m->Next) if (m->Life>0) func(m);
	}
	template<class M, class F> void Apply(F func) {
		for (CMover* m=Active->Next; m!=Active; m=m->Next) if (m->Life>0) func((M*)m);
	}
	CMover* Prev(CMover* mover) {
		CMover* prev=mover->Prev;
		return prev!=Active?prev:prev->Prev;
	}
	CMover* Next(CMover* mover) {
		CMover* next=mover->Next;
		return next!=Active?next:next->Next;
	}
};

template<class M> M* New(CVector position, CVector velocity=CVector()) {
	M* m=M::MoverList().template New<M>();
	m->Position=position, m->Velocity=velocity;
	m->Reset();
	return m;
}
template<class M> M* New(int type=0, CVector position=CVector(), CVector velocity=CVector()) {
	M* m=M::MoverList().template New<M>();
	m->Type=type, m->Position=position, m->Velocity=velocity;
	m->Reset();
	return m;
}
template<class M> void NewNWay(CVector position, CVector velocity, float angle, int count, int type=0) {
	if (count>1) for (int i=0; i<count; i++) {
		M* m=M::MoverList().template New<M>();
		m->Type=type, m->Position=position;
		m->Velocity=velocity*RotationZ(-angle*0.5f+angle*i/(count-1));
		m->Reset();
	} else New<M>(type, position, velocity);
}
template<class M> void NewCircle(CVector position, CVector velocity, int count, int type=0) {
	if (count>1) for (int i=0; i<count; i++) {
		M* m=M::MoverList().template New<M>();
		m->Type=type, m->Position=position;
		m->Velocity=velocity*RotationZ((float)i/count);
		m->Reset();
	} else New<M>(type, position, velocity);
}

